%{
#include <stdio.h>
#include <string.h>
#include "wdif.h"
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
 double sumdif = 0.0;
 double tol = 0.001;
 int times = 0;
 int maxtimes=20;

 void yyerror(char *s);
 int yylex(void);

 char *fname[2]={0,0};
 FILE *file[2];

void printsumdif(void)
  {
   printf("wdif: sumdif is %g # difs is %d\n", sumdif,times);
  }

 void checktimes(void)
   {
     if (maxtimes)
       if (times >= maxtimes) 
	 {
	   printf("wdif ends because of too many diffs\n");
	   printsumdif();
	   exit(1);
	 }
   }

 void wcompare(char *a, char *b)
   {
      if (strcmp(a,b) == 0)
         return;
       printf("word diff in lines %d,%d: '%s' and '%s'\n",
            lineno[0],lineno[1],a,b);
       times++;
       checktimes();
       return;
   }

 void ncompare(char * a, char * b)
   {
     double x,y,s;
     x=atof(a);
     if (isinf(x)) 
       return;
     y=atof(b);
     if (isinf(y)) 
       return;
      s = fabs(x-y);
      if (isnan(s))
        return;
      sumdif += s;
      if (fabs(x-y) > tol)
        {
          printf("numerical diff in lines %d,%d: '%s' and '%s' \n",
             lineno[0],lineno[1],a,b);
          times++;
          checktimes();
         }
    }


void endoffile(int f)
 {
   printf("wdif: line %d: premature and of file '%s'.\n",
          lineno[f],fname[f]);
   printsumdif();
   exit(1);
 }


%}

%union{
   char * number;
   char * word;
  }
%token  <number> NUMBER
%token  <word>   WORD
%token ENDF

%%

input:	/*empty */
	| input wordpair
	| input numberpair
	| input mixedpair
	| input bothend
	| input end0
	| input end1
	;

wordpair:	WORD WORD { wcompare($1,$2) ;   free($1); free($2); }
	;

numberpair:	NUMBER NUMBER { ncompare($1,$2) ; free($1); free($2);  }
	;

mixedpair:	NUMBER WORD { wcompare($1,$2); free($1); free($2); }
	|	WORD NUMBER { wcompare($1,$2); free($1); free($2); }
	;

bothend:	ENDF ENDF { printsumdif(); 
                            if(times == 0)
			      exit(0);
			    else
			      exit(1);}
	;

anyword:	WORD  { free($1); }
	|	NUMBER { free($1); }
	;

end0:		ENDF anyword { endoffile(0); }
	;

end1	:	anyword ENDF { endoffile(1); }
	;


%%

void help(void)
  {
    puts("wdif compares two files on a word-by-word basis.");
    puts("Differences in words are printed to stdout");
    puts("If the words seem to be numbers, a conversion is done");
    puts("to doubles, subsequentely these doubles are compared.");
    puts("White space, newlines and commas are ignored.");
    puts("In order to facilitate comparing fortran output files,");
    puts("things as 4*1.1 are expanded to 1.1 1.1 1.1 1.1");
    puts("At the end, wdif prints the sum of absolute differences");
    puts(" and the number of differences found");
    puts("");
    puts("Usage: wdif [-n N] [-t tol] file1 file2");
    puts("where:");
    puts("   N is the maximum number of differences allowed before");
    puts("     the program stops, default 20. 0 means: no stop.");
    puts("   tol is the threshold for noticing differences between numbers");
    puts("     default: 0.001");
    puts("   file1, file2: files to be compared.");
    puts("");
    puts("Example:");
    puts("  wdif -n 0 -t 0.1 old new");
  }

void errorend(void)
  {
     fprintf(stderr,"wdif: usage: wdif [-n N] [-t tol] file1 file1\n");
     exit(1);
  }

int main(int argc, char *argv[])
  {
     int i,c,err=0;
     extern char *optarg;
     extern int optind;
     char *p;

     if (argc == 1)
       {
          help();
          exit(0);
       }
     while ((c = getopt(argc, argv, "h?n:t:")) != -1)
        switch(c)
          {
            case 'h':
              help();
              exit(0);
              break;
            case 't':
              tol = strtod(optarg,&p);
              if (strlen(optarg) != (int)(p-optarg))
                {
                  fprintf(stderr,"wdif: invalid real after -t\n");
                  err++;
                }
               break;
             case 'n':
               maxtimes = strtol(optarg,&p,10);
               if (strlen(optarg) != (int)(p-optarg))
                 {
                    fprintf(stderr,"wdif: invalid integer after -n\n");
                    err++;
                 }
                break;
              case '?':
                err++;
              break;
            }
     if (argc-optind !=2)
        err++;
     if (err)
         errorend(); 
     i=0;
     for ( ; optind < argc; optind++)
       {
          fname[i]=argv[optind];
         file[i]=fopen(fname[i],"r");
         if (file[i] == 0)
           {
             fprintf(stderr,"wdif: Cannot open %s\n",fname[i]);
             errorend();
           }
          i++;
        }
     yyparse();
     return 0;
   }

void yyerror(char *s)
 {
   printf("Error %s\n",s);
 }
